/* Use always white background */
body {
  background : white;
  color      : black;
  font-family: monospace;
}

/* Show hyperlinks without text decoration, but in light yellow */
a:visited, a:link, a:active {
  text-decoration: none;
  background     : lightyellow;
}

/* Line numbers */
.linenumbers {
  width        : 40px;
  text-align   : right;
  color        : grey;
  padding-right: 10px;
  border-right : 1px solid grey;
}

/* Source code */
.sourcecode  {
  padding-left: 10px;
}

/* Code highlighting */
.pragma  { color : green     }
.comment { color : green     }
.keyword { color : blue      }
.symbol  { color : red       }
.type    { color : orange    }
.cons    { color : magenta   }
.label   { color : darkgreen }
.func    { color : purple    }
.ident   { color : black     }
.module  { color : brown     }
.number  { color : teal      }
.string  { color : maroon    }
.char    { color : maroon    }
